#pragma once

#include <cstdarg>
#include <cstdint>
#include <cstdlib>
#include <ostream>
#include <new>

namespace ChafKitCore {

enum class CallbackKind {
  ChafKitInitialized,
  ChatMessageReceived,
  LobbyCreated,
  LobbyUpdated,
  LobbyJoinRequested,
  LobbyJoined,
  LobbySearch,
  ExternalTokenIssued,
#if defined(CHAFKIT_TEST)
  TestCallback = -1,
#endif
};

enum class ChafKitResult {
  Ok = 0,
  InvalidParameters = 2,
  NotFound = 3,
};

enum class ComparisonOp {
  Equal,
  NotEqual,
  GreaterThan,
  GreaterThanOrEqual,
  LessThan,
  LessThanOrEqual,
};

enum class LobbyType : uint32_t {
  Private = 0,
  FriendsOnly = 1,
  Public = 2,
  Invisible = 3,
};

struct LobbyMember;

struct LobbySearchContext;

struct LobbySearchResult;

struct Profile;

struct CallbackHandle {
  int64_t _0;
};

struct ChafKitInitialized {
  const char *error;
};

struct ChatMessageReceived {
  const char *sender_id;
  int32_t chat_id;
};

struct LobbyCreated {
  const char *error;
  const char *lobby_id;
};

struct LobbyJoinRequested {
  const char *lobby_id;
};

struct LobbyJoined {
  const char *error;
  const char *lobby_id;
};

struct LobbyUpdated {
  const char *lobby_id;
};

struct LobbySearch {
  const char *error;
  const LobbySearchResult *search_handle;
};

struct ExternalTokenIssued {
  const char *error;
  const char *token;
};

union CallbackResultUnion {
  ChafKitInitialized chafkit_initialized;
  ChatMessageReceived chat_message_received;
  LobbyCreated lobby_created;
  LobbyJoinRequested lobby_join_requested;
  LobbyJoined lobby_joined;
  LobbyUpdated lobby_updated;
  LobbySearch lobby_search;
  ExternalTokenIssued external_token_issued;
#if defined(CHAFKIT_TEST)
  TestCallback test_callback
#endif
  ;
};

using CCallback = void(*)(const CallbackResultUnion *result, CallbackHandle handle);

struct ChafKitNetworkingMessage {
  int16_t channel;
  char *sender_id;
  uint8_t *data;
  int data_size;
};

extern "C" {

void chafkit_overlay_open();

void chafkit_overlay_close();

/// This function returns the visibility of the overlay.
///
/// # Safety
///
/// This function is `unsafe` because it dereferences a raw pointer and can therefore
/// cause undefined behavior if the pointer is null or points to invalid memory.
///
/// Before calling this function, you must ensure that the `result` pointer is non-null
/// and points to valid memory that is writable.
///
/// # Arguments
///
/// * `result` - A pointer to a `bool` value where the visibility will be stored.
///
/// # Returns
///
/// If the visibility could be retrieved and stored in the `result` pointer, returns `0`.
/// Otherwise, returns `-1`.
int chafkit_overlay_get_visibility(bool *result);

ChafKitResult chafkit_auth_get_profile(const Profile **out_profile);

CallbackHandle chafkit_auth_issue_external_token();

CallbackHandle chafkit_initialize(const char *game_id);

bool chafkit_shutdown();

CallbackHandle chafkit_callback_register(CallbackKind id, CCallback cb);

void chafkit_callback_handle_set_callback(CallbackHandle cb_handle, CCallback cb);

void chafkit_callback_unregister(CallbackHandle handle);

const char *chafkit_config_get_game_id();

CallbackHandle chafkit_matchmaking_create_lobby(const char *name,
                                                uint32_t max_members,
                                                LobbyType lobby_type,
                                                bool use_chaf_relay);

CallbackHandle chafkit_matchmaking_join_lobby(const char *lobby_id);

ChafKitResult chafkit_matchmaking_leave_lobby(const char *lobby_id);

ChafKitResult chafkit_matchmaking_destroy_lobby(const char *lobby_id);

ChafKitResult chafkit_matchmaking_invite_friend_to_lobby(const char *lobby_id, const char *user_id);

ChafKitResult chafkit_matchmaking_get_lobby_chat_entry(int32_t chat_id,
                                                       const char **out_sender_id,
                                                       const uint8_t **out_data,
                                                       int32_t *out_data_size);

ChafKitResult chafkit_matchmaking_send_lobby_chat(const uint8_t *data, int32_t data_size);

ChafKitResult chafkit_lobby_get_name(const char *lobby_id, const char **out_name);

ChafKitResult chafkit_lobby_get_max_members(const char *lobby_id, unsigned int *out_max_members);

ChafKitResult chafkit_lobby_get_type(const char *lobby_id, LobbyType *out_lobby_type);

ChafKitResult chafkit_lobby_get_data(const char *lobby_id,
                                     const char *field,
                                     const char **out_data);

ChafKitResult chafkit_lobby_get_original_host_id(const char *lobby_id,
                                                 const char **out_original_host_id);

ChafKitResult chafkit_lobby_get_current_host_id(const char *lobby_id,
                                                const char **out_current_host_id);

ChafKitResult chafkit_lobby_get_members_len(const char *lobby_id, unsigned int *out_members_len);

ChafKitResult chafkit_lobby_get_use_chaf_relay(const char *lobby_id, bool *out_use_chaf_relay);

ChafKitResult chafkit_lobby_get_is_joinable(const char *lobby_id, bool *out_is_joinable);

ChafKitResult chafkit_lobby_get_created_at(const char *lobby_id, const char **out_created_at);

ChafKitResult chafkit_lobby_get_id(const char *lobby_id, const char **out_id);

ChafKitResult chafkit_lobby_set_data(const char *lobby_id, const char *field, const char *value);

ChafKitResult chafkit_lobby_delete_data(const char *lobby_id, const char *field);

ChafKitResult chafkit_lobby_get_game_server(const char *lobby_id,
                                            const char **out_ip,
                                            int *out_port,
                                            const char **out_server_id);

ChafKitResult chafkit_lobby_set_game_server(const char *lobby_id,
                                            const char *ip,
                                            int port,
                                            const char *server_id);

ChafKitResult chafkit_lobby_set_joinable(const char *lobby_id, bool is_joinable);

ChafKitResult chafkit_lobby_set_max_members(const char *lobby_id, int32_t max_members);

ChafKitResult chafkit_lobby_change_host(const char *lobby_id, const char *member_id);

ChafKitResult chafkit_lobby_set_type(const char *lobby_id, LobbyType lobby_type);

ChafKitResult chafkit_lobby_refresh(const char *lobby_id);

ChafKitResult chafkit_lobby_set_member_data(const char *lobby_id,
                                            const char *field,
                                            const char *value);

ChafKitResult chafkit_lobby_get_member_clone(const char *lobby_id,
                                             unsigned int idx,
                                             const LobbyMember **out_lobby_member);

ChafKitResult chafkit_lobby_member_get_profile(const LobbyMember *lobby_member,
                                               const Profile **out_profile);

ChafKitResult chafkit_lobby_member_get_joined_at(const LobbyMember *lobby_member,
                                                 const char **out_joined_at);

ChafKitResult chafkit_lobby_member_get_data(const LobbyMember *lobby_member,
                                            const char *field,
                                            const char **out_value);

ChafKitResult chafkit_lobby_member_release(LobbyMember *lobby_member);

ChafKitResult chafkit_lobby_search_result_get_by_index(const LobbySearchResult *search_handle,
                                                       int index,
                                                       const char **out_lobby_id);

ChafKitResult chafkit_lobby_search_result_len(LobbySearchResult *search_handle, int *out_len);

ChafKitResult chafkit_lobby_search_create(const LobbySearchContext **out_handle);

ChafKitResult chafkit_lobby_search_release(LobbySearchContext *handle);

ChafKitResult chafkit_lobby_add_filter_slots_available(LobbySearchContext *handle,
                                                       int32_t slots_available);

ChafKitResult chafkit_lobby_search_add_distance_filter(LobbySearchContext *handle,
                                                       int32_t distance_level);

ChafKitResult chafkit_lobby_search_add_string_filter(LobbySearchContext *handle,
                                                     const char *field,
                                                     const char *value_to_match,
                                                     ComparisonOp comparison_op);

ChafKitResult chafkit_lobby_search_add_numerical_filter(LobbySearchContext *handle,
                                                        const char *field,
                                                        int value_to_match,
                                                        ComparisonOp comparison_op);

ChafKitResult chafkit_lobby_search_remove_filter(LobbySearchContext *handle,
                                                 const char *field,
                                                 ComparisonOp comparison_op);

ChafKitResult chafkit_lobby_search_order_by_near(LobbySearchContext *handle,
                                                 const char *field,
                                                 int32_t value);

ChafKitResult chafkit_lobby_search_set_result_page(LobbySearchContext *handle,
                                                   int offset,
                                                   int limit);

CallbackHandle chafkit_matchmaking_search_find(const LobbySearchContext *handle);

ChafKitResult chafkit_networking_connect();

ChafKitResult chafkit_networking_close();

ChafKitResult chafkit_networking_send_message(void *netgroup,
                                              short channel,
                                              const char *receiver_id,
                                              const uint8_t *data,
                                              int data_size);

/// Receives messages from a network group, and stores them in an array.
///
/// # Safety
///
/// This function is safe as long as:
///
/// - `netgroup` points to a valid memory location that can be cast to a `Box<dyn NetworkingGroup>`.
/// - `out_messages` points to an array of pointers to `ChafKitNetworkingMessage` structures that
///   is at least `max_messages` in length.
///
/// Failure to meet these conditions will cause undefined behavior.
///
/// # Arguments
///
/// - `netgroup`: A raw pointer to a NetworkingGroup object. Must be a valid memory location.
/// - `channel`: The channel to receive messages from.
/// - `out_messages`: A raw pointer to the first element in an array of raw pointers.
///   This array will be populated with the received messages.
/// - `max_messages`: The maximum number of messages to receive.
///
/// # Returns
///
/// The number of messages actually received, as a C integer.
int chafkit_networking_receive_messages(void *netgroup,
                                        short channel,
                                        ChafKitNetworkingMessage **out_messages,
                                        int max_messages);

ChafKitResult chafkit_networking_accept_user(void *netgroup, const char *user_id);

/// Releases a `ChafKitNetworkingMessage` previously allocated by
/// `chafkit_networking_receive_messages`.
///
/// # Safety
///
/// This function is safe as long as `message` is a valid pointer to a `ChafKitNetworkingMessage`
/// structure that has not been previously released.
///
/// Once a `ChafKitNetworkingMessage` has been released, it should not be used again. Attempting
/// to release it again or use it in any way will cause undefined behavior.
///
/// # Arguments
///
/// - `message`: A raw pointer to a `ChafKitNetworkingMessage` that was previously allocated.
///
/// # Returns
///
/// A `ChafKitResult` enum indicating the success or failure of the operation. If the operation
/// was successful, it returns `ChafKitResult::Ok`. If `message` is null or points to invalid
/// memory, it returns `ChafKitResult::InvalidParameters`.
ChafKitResult chafkit_release_networking_message(ChafKitNetworkingMessage *message);

ChafKitResult chafkit_networking_create_netgroup_by_lobby(const char *lobby_id,
                                                          void **out_netgroup);

ChafKitResult chafkit_release_netgroup(void *netgroup);

ChafKitResult chafkit_profile_get_username(const Profile *profile, const char **out_username);

ChafKitResult chafkit_profile_get_display_name(const Profile *profile,
                                               const char **out_display_name);

ChafKitResult chafkit_profile_get_user_id(const Profile *profile, const char **out_user_id);

ChafKitResult chafkit_profile_get_photo(const Profile *profile, const char **out_photo);

} // extern "C"

} // namespace ChafKitCore
